"""Module to handle Image Datasets."""

from typing import Union

from .bands import Band, Bit, BitBand, BitGroup, Bitmask, CategoricalBand, Category, OpticalBand
from .custom_types import ListNamespace
from .dataset import Dataset


class Image(Dataset):
    def __init__(self, href: str, name: str, parent):
        """Image Dataset."""
        super(Image, self).__init__(href=href, name=name, parent=parent)
        self._bands: ListNamespace[Union[Band, BitBand, CategoricalBand, OpticalBand]] = (
            ListNamespace(key="name")
        )

    @property
    def bands(self) -> ListNamespace[Union[Band, BitBand, CategoricalBand, OpticalBand]]:
        """Image bands."""
        if len(self._bands) == 0:
            bands = []
            summ = self._get_summaries()
            eobands = summ.get("eo:bands", [])
            general_scale = summ.get("gsd")
            b: Union[Band, BitBand, CategoricalBand, OpticalBand]  # Explicit type hint for 'b'
            if isinstance(general_scale, (list, tuple)):
                # TODO: check/investigate why it's a list and what happens if it has more than 1 item
                general_scale = general_scale[0]
            for band in eobands:
                name = band.get("name")
                scale = band.get("gsd", general_scale)
                desc = band.get("description")
                range = self._get_summaries().get(name, {})
                if "gee:wavelength" in band or "center_wavelength" in band:
                    # OpticalBand
                    center_wl = band.get("center_wavelength")
                    offset = band.get("offset", 0)
                    multiplier = band.get("scale", 0)
                    b = OpticalBand(name, scale, desc, center_wl, offset, multiplier, range)
                elif "gee:classes" in band:
                    # CategoricalBand
                    classes = band.get("gee:classes")
                    cat_list = []
                    for cls in classes:
                        cat = Category(cls["value"], cls["description"], cls.get("color"))
                        cat_list.append(cat)
                    class_info = ListNamespace(*cat_list, key="description")
                    b = CategoricalBand(name, scale, desc, class_info)
                elif "gee:bitmask" in band:
                    # BitBand
                    bitmask_raw = band.get("gee:bitmask")
                    parts = bitmask_raw["bitmask_parts"]
                    bit_groups = []
                    for part in parts:
                        bits = []
                        values = part.get("values")
                        if not values:
                            continue
                        for value in values:
                            bit = Bit(value["value"], value["description"])
                            bits.append(bit)
                        bitsns = ListNamespace(*bits, key="description")
                        group = BitGroup(
                            part["first_bit"], part["bit_count"], part["description"], bitsns
                        )
                        bit_groups.append(group)
                    partsns = ListNamespace(*bit_groups, key="description")
                    bitmask = Bitmask(partsns, bitmask_raw["total_bit_count"])
                    b = BitBand(name, scale, desc, bitmask)
                else:
                    b = Band(name, scale, desc, range)
                bands.append(b)
            self._bands = ListNamespace(*bands, key="name")
        return self._bands



"""Test the geestac catalog module."""

from geestac import eecatalog, fromId

cop = eecatalog.COPERNICUS()


class TestEECatalog:
    """Test EECatalog."""

    def test_data(self):
        """Test eecatalog data."""
        assert cop.name == "COPERNICUS"
        assert isinstance(cop.data, dict)

    def test_children(self):
        """Test catalog children."""
        children = cop.children.as_list("name")
        assert children == [
            "CORINE_V18_5_1_100m",
            "CORINE_V20_100m",
            "Landcover_100m_Proba_V_C3_Global",
            "Landcover_100m_Proba_V_Global",
            "S1_GRD",
            "S2",
            "S2_HARMONIZED",
            "S2_CLOUD_PROBABILITY",
            "S2_SR",
            "S2_SR_HARMONIZED",
            "S3_OLCI",
            "S5P_NRTI_L3_AER_AI",
            "S5P_NRTI_L3_AER_LH",
            "S5P_NRTI_L3_CLOUD",
            "S5P_NRTI_L3_CO",
            "S5P_NRTI_L3_HCHO",
            "S5P_NRTI_L3_NO2",
            "S5P_NRTI_L3_O3",
            "S5P_NRTI_L3_SO2",
            "S5P_OFFL_L3_AER_AI",
            "S5P_OFFL_L3_AER_LH",
            "S5P_OFFL_L3_CH4",
            "S5P_OFFL_L3_CLOUD",
            "S5P_OFFL_L3_CO",
            "S5P_OFFL_L3_HCHO",
            "S5P_OFFL_L3_NO2",
            "S5P_OFFL_L3_O3",
            "S5P_OFFL_L3_O3_TCL",
            "S5P_OFFL_L3_SO2",
            "DEM_GLO30",
        ]

    def test_parent(self):
        """Test parent."""
        assert cop.parent == eecatalog

    def test_from_id_catalog(self):
        """Test fromId (catalog)."""
        assert fromId("COPERNICUS") == cop

    def test_from_id_dataset(self):
        """Test fromId (dataaset)."""
        assert fromId("COPERNICUS/S2_SR_HARMONIZED") == cop.S2_SR_HARMONIZED()
        assert fromId("LANDSAT/LC09/C02/T1") == eecatalog.LANDSAT.LC09_C02_T1()



"""Helpers for this package."""


def format_attribute(attribute: str) -> str:
    """Convert characters from any attribute to match the format of class/object attributes.

    Args:
        attribute: an attribute.
    """
    to_replace = "- ."
    for char in to_replace:
        attribute = attribute.replace(char, "_")
    return attribute



"""Module to handle FeatureCollection Datasets."""

from .dataset import Dataset


class FeatureCollection(Dataset):
    def __init__(self, href: str, name: str, parent):
        """Feature Collection."""
        super(FeatureCollection, self).__init__(href, name, parent)



"""Test the geestac catalog module."""

import ee

from geestac import eecatalog


class TestImage:
    """Test an optical EO ImageCollection."""

    srtm = eecatalog.CGIAR().SRTM90_V4()

    def test_ee_type(self):
        """Test an optical earth observation ImageCollection."""
        assert self.srtm.eeType == ee.Image



"""Test custom types."""

from geestac.custom_types import ListNamespace


class Tree:
    """Tree class for testing purposes."""

    def __init__(self, genus: str, specie: str):
        """Plant a tree."""
        self.genus = genus
        self.specie = specie


class TestListNamepace:
    """Test ListNamespace."""

    lenga = Tree("Nothofagus", "pumilo")
    laura = Tree("Schinus", "patagonicus")
    ln = ListNamespace(lenga, laura, key="genus")

    def test_as_list_default(self):
        """Test as_list with default params."""
        assert self.ln.as_list() == [self.lenga, self.laura]

    def test_as_list_key(self):
        """Test as_list with key param."""
        assert self.ln.as_list("genus") == ["Nothofagus", "Schinus"]

    def test_as_dict_default(self):
        """Test as_dict with default param."""
        assert self.ln.as_dict() == {"Nothofagus": self.lenga, "Schinus": self.laura}

    def test_as_dict_value(self):
        """Test as_dict with value param."""
        assert self.ln.as_dict("specie") == {"Nothofagus": "pumilo", "Schinus": "patagonicus"}



"""STAC Base class."""

import requests


class STAC:
    """Base STAC class."""

    def __init__(self, href: str, name: str, parent=None):
        """Base STAC class.

        Fetch the STAC data at init.

        Args:
            href: URL of the catalog / dataset.
            name: name of the catalog / dataset.
            parent: parent STAC class.
        """
        self.href = href
        self.name = name
        self.data: dict = {}
        self.parent = parent
        self._lazy = True

    def is_lazy(self) -> bool:
        """Lazy means it has not fetched the real data yet."""
        return self._lazy

    def __repr__(self):
        """Object representation."""
        return f"{self.name}{' (lazy)' if self.is_lazy() else ''}"

    @property
    def description(self):
        """Description of the Catalog."""
        return self.data.get("description")

    @property
    def version(self):
        """STAC Version."""
        return self.data.get("stac_version")

    def __call__(self):
        """Fetch the data."""
        if self.is_lazy():
            self.data = requests.get(self.href).json()
            self._lazy = False
        return self

    def __eq__(self, other):
        """Compare with another object."""
        if not isinstance(other, STAC):
            return False
        data = self.data == other.data
        name = self.name == other.name
        return data and name



"""Test the geestac catalog module."""

import ee

from geestac import eecatalog

fao = eecatalog.FAO().GAUL_2015_level0()


class TestFeatureCollection:
    """Test an optical EO ImageCollection."""

    def test_ee_type(self):
        """Test an optical earth observation ImageCollection."""
        assert fao.eeType == ee.FeatureCollection

    def test_properties(self):
        """Test properties."""
        assert fao.properties.as_list("name") == [
            "ADM0_CODE",
            "ADM0_NAME",
            "DISP_AREA",
            "STATUS",
            "Shape_Area",
            "Shape_Leng",
            "EXP0_YEAR",
            "STR0_YEAR",
        ]



"""Module to handle GEE Datasets."""

import ee

from .custom_types import ListNamespace
from .stac import STAC


class Property:
    """A class that represents a property."""

    def __init__(self, name: str, type: str, description: str):
        """An object that represents a property.

        Args:
            name: name of the property.
            type: data type of the property.
            description: description of the property.
        """
        self.name = name
        self.type = type
        self.description = description

    def __repr__(self) -> str:
        """Object representation."""
        return f"{self.name} ({self.type})"


class Dataset(STAC):
    def __init__(self, href: str, name: str, parent):
        """Dataset class.

        Args:
            href: the URL.
            name: name of the Dataset.
            parent: the parent Catalog.
        """
        super(Dataset, self).__init__(href, name, parent)

    def __repr__(self) -> str:
        """Object representation."""
        eetype = f" ({self.eeType.name()})" if self.eeType else ""
        status = f" ({self.status})" if self.status else ""
        return f"{self.name}{eetype}{status}"

    def _get_interval(self) -> list:
        """Temporal interval."""
        extent = self.data.get("extent", {})
        temporal_resolution = extent.get("temporal", {})
        interval = temporal_resolution.get("interval", [[None, None]])
        return interval

    def _get_summaries(self) -> dict:
        return self.data.get("summaries", {})

    @property
    def start_date(self) -> str | None:
        """Start date of the dataset."""
        try:
            extent = self.data.get("extent", {})
            temporal_resolution = extent.get("temporal", {})
            interval = temporal_resolution.get("interval", [[None, None]])
            return interval[0][0]
        except KeyError:
            return None

    @property
    def end_date(self) -> str | None:
        """End date of the dataset."""
        try:
            return self._get_interval()[0][1]
        except KeyError:
            return None

    @property
    def spatial_extent(self) -> ee.Geometry | None:
        """Spatial Extent."""
        extent = self.data.get("extent", {})
        spatial = extent.get("spatial", {})
        bbox = spatial.get("bbox", None)
        return bbox

    @property
    def eeType(self):
        """Earth Engine Object Type."""
        ty = self.data.get("gee:type")
        types = {
            "table": ee.FeatureCollection,
            "image": ee.Image,
            "image_collection": ee.ImageCollection,
        }
        return types.get(ty)

    @property
    def license(self) -> str | None:
        """License of Use."""
        return self.data.get("license")

    @property
    def assetId(self) -> str | None:
        """Earth Engine Asset Id."""
        return self.data.get("id")

    @property
    def eeObject(self):
        """Earth Engine Object."""
        return self.eeType(self.assetId) if self.assetId else None

    @property
    def status(self) -> str | None:
        """Status of the Dataset."""
        return self.data.get("gee:status")

    @property
    def properties(self) -> ListNamespace:
        """Dataset properties."""
        summ = self.data.get("summaries", {})
        schema = summ.get("gee:schema", [])
        prop = []
        for sch in schema:
            prop.append(Property(sch.get("name"), sch.get("type"), sch.get("description")))
        return ListNamespace(*prop, key="name")

    @property
    def terms_of_use(self) -> str | None:
        """Terms of use."""
        return self.data.get("gee:terms_of_use")

    @property
    def DOI(self) -> str | None:
        """DOI."""
        return self.data.get("sci:doy")

    @property
    def citation(self) -> str | None:
        """Citation."""
        return self.data.get("sci:citation")



"""Test the geestac catalog module."""

from types import SimpleNamespace

import ee

from geestac import eecatalog

l9 = eecatalog.LANDSAT().LC09_C02_T1()
s2 = eecatalog.COPERNICUS().S2_SR()


class TestOpticalImageCollection:
    """Test an optical EO ImageCollection."""

    def test_ee_type(self):
        """Test ee type."""
        assert l9.eeType == ee.ImageCollection

    def test_start_date(self):
        """Test start date."""
        assert l9.start_date == "2021-10-31T00:00:00Z"

    def test_status(self):
        """Test status."""
        assert s2.status == "deprecated"

    def test_asset_id(self):
        """Test AssetId."""
        assert s2.assetId == "COPERNICUS/S2_SR"

    def test_properties(self):
        """Test Properties."""
        properties = s2.properties.as_list("name")
        assert properties == [
            "AOT_RETRIEVAL_ACCURACY",
            "CLOUDY_PIXEL_PERCENTAGE",
            "CLOUD_COVERAGE_ASSESSMENT",
            "CLOUDY_SHADOW_PERCENTAGE",
            "DARK_FEATURES_PERCENTAGE",
            "DATASTRIP_ID",
            "DATATAKE_IDENTIFIER",
            "DATATAKE_TYPE",
            "DEGRADED_MSI_DATA_PERCENTAGE",
            "FORMAT_CORRECTNESS",
            "GENERAL_QUALITY",
            "GENERATION_TIME",
            "GEOMETRIC_QUALITY",
            "GRANULE_ID",
            "HIGH_PROBA_CLOUDS_PERCENTAGE",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B1",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B2",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B3",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B4",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B5",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B6",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B7",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B8",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B8A",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B9",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B10",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B11",
            "MEAN_INCIDENCE_AZIMUTH_ANGLE_B12",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B1",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B2",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B3",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B4",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B5",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B6",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B7",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B8",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B8A",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B9",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B10",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B11",
            "MEAN_INCIDENCE_ZENITH_ANGLE_B12",
            "MEAN_SOLAR_AZIMUTH_ANGLE",
            "MEAN_SOLAR_ZENITH_ANGLE",
            "MEDIUM_PROBA_CLOUDS_PERCENTAGE",
            "MGRS_TILE",
            "NODATA_PIXEL_PERCENTAGE",
            "NOT_VEGETATED_PERCENTAGE",
            "PROCESSING_BASELINE",
            "PRODUCT_ID",
            "RADIATIVE_TRANSFER_ACCURACY",
            "RADIOMETRIC_QUALITY",
            "REFLECTANCE_CONVERSION_CORRECTION",
            "SATURATED_DEFECTIVE_PIXEL_PERCENTAGE",
            "SENSING_ORBIT_DIRECTION",
            "SENSING_ORBIT_NUMBER",
            "SENSOR_QUALITY",
            "SOLAR_IRRADIANCE_B1",
            "SOLAR_IRRADIANCE_B2",
            "SOLAR_IRRADIANCE_B3",
            "SOLAR_IRRADIANCE_B4",
            "SOLAR_IRRADIANCE_B5",
            "SOLAR_IRRADIANCE_B6",
            "SOLAR_IRRADIANCE_B7",
            "SOLAR_IRRADIANCE_B8",
            "SOLAR_IRRADIANCE_B8A",
            "SOLAR_IRRADIANCE_B9",
            "SOLAR_IRRADIANCE_B10",
            "SOLAR_IRRADIANCE_B11",
            "SOLAR_IRRADIANCE_B12",
            "SNOW_ICE_PERCENTAGE",
            "SPACECRAFT_NAME",
            "THIN_CIRRUS_PERCENTAGE",
            "UNCLASSIFIED_PERCENTAGE",
            "VEGETATION_PERCENTAGE",
            "WATER_PERCENTAGE",
            "WATER_VAPOUR_RETRIEVAL_ACCURACY",
        ]

    def test_revisit(self):
        """Test revisit."""
        revisit = l9.revisit
        assert isinstance(revisit, SimpleNamespace)
        assert revisit.interval == 16
        assert revisit.unit == "day"


class TestBands:
    """Test bands."""

    def test_band_names(self):
        """Test band names."""
        names = s2.bands.as_list("name")
        assert names == [
            "B1",
            "B2",
            "B3",
            "B4",
            "B5",
            "B6",
            "B7",
            "B8",
            "B8A",
            "B9",
            "B11",
            "B12",
            "AOT",
            "WVP",
            "SCL",
            "TCI_R",
            "TCI_G",
            "TCI_B",
            "MSK_CLDPRB",
            "MSK_SNWPRB",
            "QA10",
            "QA20",
            "QA60",
            "MSK_CLASSI_OPAQUE",
            "MSK_CLASSI_CIRRUS",
            "MSK_CLASSI_SNOW_ICE",
        ]

    def test_band_scales(self):
        """Test band scales."""
        scales = l9.bands.as_dict("scale")
        assert scales == {
            "B1": 30,
            "B2": 30,
            "B3": 30,
            "B4": 30,
            "B5": 30,
            "B6": 30,
            "B7": 30,
            "B8": 15,
            "B9": 30,
            "B10": 30,
            "B11": 30,
            "QA_PIXEL": 30,
            "QA_RADSAT": 30,
            "SAA": 30,
            "SZA": 30,
            "VAA": 30,
            "VZA": 30,
        }

    def test_band_wavelength(self):
        """Test wavelength."""
        wavelength = s2.bands.as_dict("center_wavelength")
        assert wavelength == {
            "B1": 0.4439,
            "B2": 0.4966,
            "B3": 0.56,
            "B4": 0.6645,
            "B5": 0.7039,
            "B6": 0.7402,
            "B7": 0.7825,
            "B8": 0.8351,
            "B8A": 0.8648,
            "B9": 0.945,
            "B11": 1.6137,
            "B12": 2.2024,
        }


class TestCategoricalBand:
    """Test CategoricalBand."""

    def test_classes(self):
        """Test classes."""
        classes = s2.bands.SCL.class_info.as_dict("value")
        assert classes == {
            "Saturated or defective": 1,
            "Dark Area Pixels": 2,
            "Cloud Shadows": 3,
            "Vegetation": 4,
            "Bare Soils": 5,
            "Water": 6,
            "Clouds Low Probability / Unclassified": 7,
            "Clouds Medium Probability": 8,
            "Clouds High Probability": 9,
            "Cirrus": 10,
            "Snow / Ice": 11,
        }


class TestBitBand:
    """Test BitBand."""

    def test_bitmask(self):
        """Test the bitmask."""
        bitmask = l9.bands.QA_PIXEL.bitmask
        assert bitmask.total == 16
        assert bitmask.to_dict() == {
            "0-0-Fill": {"0": "Image data", "1": "Fill data"},
            "1-1-Dilated Cloud": {"0": "Cloud is not dilated or no cloud", "1": "Cloud dilation"},
            "2-2-Cirrus": {
                "0": "Cirrus Confidence: no confidence level set or Low Confidence",
                "1": "High confidence cirrus",
            },
            "3-3-Cloud": {"0": "Cloud confidence is not high", "1": "High confidence cloud"},
            "4-4-Cloud Shadow": {
                "0": "Cloud Shadow Confidence is not high",
                "1": "High confidence cloud shadow",
            },
            "5-5-Snow": {"0": "Snow/Ice Confidence is not high", "1": "High confidence snow cover"},
            "6-6-Clear": {
                "0": "Cloud or Dilated Cloud bits are set",
                "1": "Cloud and Dilated Cloud bits are not set",
            },
            "7-7-Water": {"0": "Land or cloud", "1": "Water"},
            "8-9-Cloud Confidence": {
                "0": "No confidence level set",
                "1": "Low confidence",
                "2": "Medium confidence",
                "3": "High confidence",
            },
            "10-11-Cloud Shadow Confidence": {
                "0": "No confidence level set",
                "1": "Low confidence",
                "2": "Reserved",
                "3": "High confidence",
            },
            "12-13-Snow / Ice Confidence": {
                "0": "No confidence level set",
                "1": "Low confidence",
                "2": "Reserved",
                "3": "High confidence",
            },
            "14-15-Cirrus Confidence": {
                "0": "No confidence level set",
                "1": "Low confidence",
                "2": "Reserved",
                "3": "High confidence",
            },
        }



"""Module to handle ImageCollection Datasets."""

from types import SimpleNamespace

from .image import Image


class ImageCollection(Image):
    def __init__(self, href: str, name: str, parent):
        """ImageCollection Dataset."""
        super(ImageCollection, self).__init__(href=href, name=name, parent=parent)

    @property
    def revisit(self) -> SimpleNamespace:
        """Revisit time.

        Returns:
            a dict as follows: {'interval', 'unit'}
        """
        interval = self.data.get("gee:interval", {})
        if interval.get("type") == "revisit_interval":
            return SimpleNamespace(
                **{"interval": interval.get("interval"), "unit": interval.get("unit")}
            )
        return interval



"""Earth Engine STAC Catalog."""

import re
from typing import Union

import requests

from .custom_types import ListNamespace
from .dataset import Dataset
from .feature_collection import FeatureCollection
from .image import Image
from .image_collection import ImageCollection
from .stac import STAC


class LazyDataset(STAC):
    def __init__(self, href: str, name: str, parent):
        """Catalog."""
        super(LazyDataset, self).__init__(href, name, parent)
        self.data = {}

    def __call__(self):
        """Fetch data and return the corresponding object."""
        super(LazyDataset, self).__call__()
        eetype = self.data.get("gee:type")
        if eetype == "image":
            # fetch data here
            ds = Image(self.href, self.name, self.parent)()
        elif eetype == "image_collection":
            ds = ImageCollection(self.href, self.name, self.parent)()
        elif eetype == "table":
            ds = FeatureCollection(self.href, self.name, self.parent)()
        else:
            ds = Dataset(self.href, self.name, self.parent)()
        return ds


class Catalog(STAC):
    def __init__(self, href: str, name: str, parent):
        """Catalog."""
        super(Catalog, self).__init__(href, name, parent)
        self.data = {}
        self.children: ListNamespace[Union[Dataset, LazyDataset]] = ListNamespace(key="name")

    def __call__(self):
        """Fetch data."""
        if self.is_lazy():
            self.data = requests.get(self.href).json()
            self._get_datasets()
            self._lazy = False
        return self

    def _get_datasets(self):
        """Get all catalogs and set them as instance properties."""
        for link in self.data.get("links", []):
            if link["rel"] == "child":
                name = link["title"].replace("-", "_")
                if re.match(f"^{self.name}", name):
                    name = name.replace(f"{self.name}_", "")
                catalog = LazyDataset(link["href"], name, self)
                self.children._append(catalog)
                self.__setattr__(name, catalog)


class EECatalog(STAC):
    """Earth Engine STAC Catalog.

    This Catalog contains a set of Catalogs accessible via attributes.

    This is always the root for all children.
    """

    base_url = "https://earthengine-stac.storage.googleapis.com/catalog/catalog.json"

    def __init__(self):
        """Earth Engine STAC Catalog."""
        super(EECatalog, self).__init__(self.base_url, "EECatalog")
        # self.children = DictNamespace()
        self.children = ListNamespace(key="name")
        self.data = requests.get(self.base_url).json()
        self._get_catalogs()

    def _get_catalogs(self):
        """Get all catalogs and set them as instance properties."""
        for link in self.data.get("links", []):
            if link["rel"] == "child":
                name = link["title"].replace("-", "_")
                catalog = Catalog(link["href"], name, self)
                if re.match(f"^{self.name}", name):
                    name = name.replace(f"{self.name}_", "")
                # setattr(self.children, name, catalog)
                # self.children[name] = catalog
                self.children._append(catalog)
                self.__setattr__(name, catalog)



"""Test helpers."""

from geestac import helpers


class TestHelpers:
    """Test helpers."""

    def test_format_attribute(self):
        """Test format_attribute function."""
        attr = "hello world - 1.1"
        expected = "hello_world___1_1"
        assert helpers.format_attribute(attr) == expected



"""Module for bands."""

from .custom_types import ListNamespace


class Category:
    """Class that represents a category in a CategoricalBand."""

    def __init__(self, value: int, description: str, color: str | None = None):
        """Initialize a category.

        Args:
            value: value of the category.
            description: description of the category.
            color: color associated to the category (optional).
        """
        self.value = value
        self.description = description
        # TODO: generate a color if not given
        self.color = color


class Bit:
    """Class that represents a single bit."""

    def __init__(self, value: int, description: str):
        """Initialize a Bit.

        Args:
            value: value of the bit.
            description: description of the bit.
        """
        self.value = value
        self.description = description


class BitGroup:
    """Class that represents a bit group."""

    def __init__(self, first: int, count: int, description: str, bits: ListNamespace[Bit]):
        """Initialize a bit group.

        Args:
            first: first bit.
            count: number of bits.
            description: description of the bit information.
            bits: a ListNamespace of Bit objects.
        """
        self.first = first
        self.count = count
        self.description = description
        self.bits = bits

    def to_dict(self) -> dict:
        """Convert a bit group to a dict."""
        start = self.first
        end = self.first + self.count - 1
        bits_key = f"{start}-{end}-{self.description}"
        bits_value = {}
        for bit in self.bits:
            bits_value[str(bit.value)] = bit.description
        return {bits_key: bits_value}


class Bitmask:
    """Class that represents a bit mask in a BitBand."""

    def __init__(self, parts: ListNamespace[BitGroup], total: int):
        """Initialize a bitmask.

        Args:
            parts: a ListNamespace of BitGroup.
            total: total number of bits.
        """
        self.parts = parts
        self.total = total

    def to_dict(self) -> dict:
        """Convert a Bitmask into a dict."""
        final = {}
        for group in self.parts:
            final.update(group.to_dict())
        return final


class Band:
    def __init__(self, name: str, scale: int | float, description: str, range: dict | None = None):
        """General Band class.

        Args:
            name: name of the band.
            scale: spatial resolution.
            description: description of the band.
            range: range of the values.
        """
        self.name = name
        self.scale = scale
        self.description = description
        self._range = range if range is not None else {}

    def __repr__(self):
        """Object representation."""
        return self.name

    @property
    def minimum_value(self) -> int | float | None:
        """If exists in the STAC, return the minimum value."""
        return self._range.get("minimum")

    @property
    def maximum_value(self) -> int | float | None:
        """If exists in the STAC, return the maximum value."""
        return self._range.get("maximum")

    @property
    def estimated_range(self) -> bool | None:
        """If range exists in STAC, return whether the range is estimated or not."""
        return self._range.get("gee:estimated_range")


class OpticalBand(Band):
    def __init__(
        self,
        name: str,
        scale: int | float,
        description: str,
        center_wavelength: int | float | None = None,
        offset: int | float | None = None,
        multiplier: int | float | None = None,
        range: dict | None = None,
    ):
        """Optical Band.

        Args:
            name: name of the band.
            scale: spatial resolution.
            description: description of the band.
            center_wavelength: center wavelength.
            offset: value to "move" the raw value.
            multiplier: value to "multiply" the raw value (GEE calls this value 'scale')
            range: range of the values.
        """
        self.center_wavelength = center_wavelength
        self.offset = offset
        self.multiplier = multiplier
        super(OpticalBand, self).__init__(name, scale, description, range)


class CategoricalBand(Band):
    def __init__(
        self, name: str, scale: int | float, description: str, class_info: ListNamespace[Category]
    ):
        """Categorical Band.

        Args:
            name: name of the band.
            scale: spatial resolution.
            description: description of the band.
            class_info: a ListNamespace with the class value as keys and the
                        class names as values.
        """
        self.class_info = class_info
        super(CategoricalBand, self).__init__(name, scale, description)


class BitBand(Band):
    def __init__(self, name: str, scale: int | float, description: str, bitmask: Bitmask):
        """Bit Band (bits mask).

        Args:
            name: name of the band.
            scale: spatial resolution.
            description: description of the band.
            bitmask: the bit mask.
        """
        self.bitmask = bitmask
        super(BitBand, self).__init__(name, scale, description)



"""The init file of the package."""

__version__ = "0.0.1"
__author__ = "Rodrigo Esteban Principe"
__email__ = "fitoprincipe82@gmail.com"

from .catalog import EECatalog

eecatalog = EECatalog()


def fromId(assetId: str):
    """Load a Catalog or Dataset from an ID."""
    parts = assetId.split("/")
    if len(parts) == 1:
        return eecatalog.children.as_dict()[assetId]()
    else:
        catalog = eecatalog.children.as_dict()[parts[0]]()
        dataset = catalog.children.as_dict()["_".join(parts[1:])]()
        return dataset



"""Custom types."""

import json
import logging
from typing import Any, Generic, List, TypeVar

from .helpers import format_attribute

logger = logging.getLogger("geestac")

# Create a TypeVar for generic typing
T = TypeVar("T")


class ListNamespace(Generic[T]):
    """A Namespace to store objects with a unique attribute (key)."""

    def __init__(self, *args: T, key: str):
        """A Namespace to store objects with a unique attribute (key).

        If the value of the key is not suitable for an attribute, it'll be
        converted with format_attribute function. The values of the keys must
        be unique, else the first occurrence will be kept.

        Args:
            args: a list of objects.
            key: the attribute of the object to get the key from.
        """
        self._key = key
        self._args: List[T] = []
        for obj in args:
            self.__setattr__(getattr(obj, key), obj)

    def keys(self) -> list:
        """Key values as list."""
        return [getattr(obj, self._key) for obj in self._args]

    def __setattr__(self, key: str, value: Any):
        """Set attribute to this object."""
        if key in ["_key", "_args"]:
            super().__setattr__(key, value)
            return self
        key = format_attribute(key)
        try:
            getattr(self, key)
        except AttributeError:
            args = list(self._args)
            args.append(value)
            self._args = args
            super().__setattr__(key, value)
        else:
            raise AttributeError(f"object with attribute {self._key} = {key} already exists.")
        return self

    def as_list(self, key: str | None = None) -> list:
        """ListNamespace as list.

        Args:
            key: attribute of the object to put in the list. If None it will put the objects in the list.
        """
        if key is None:
            return list(self._args)
        else:
            args: List[Any] = []
            for arg in self._args:
                try:
                    attr = getattr(arg, key)
                except AttributeError:
                    continue
                else:
                    args.append(attr)
            return args

    def as_dict(self, value: str | None = None) -> dict:
        """ListNamespace as a dict.

        Args:
            value: attribute of the object to use as value.
        """
        keys = self.as_list(self._key)
        values = self.as_list(value)
        return dict(zip(keys, values))

    def __repr__(self):
        """Object representation."""
        return json.dumps([item for item in self.as_list(self._key)])

    def __getitem__(self, index: int | str):
        """Get item from ListNamepace.

        Args:
            index: index of the element in the ListNamespace.
        """
        if isinstance(index, int):
            return self._args[index]
        else:
            return self.as_dict()[str]

    def __len__(self):
        """Length of the ListNamespace."""
        return len(self._args)

    def _append(self, obj: T):
        """Append an object to the list."""
        key = getattr(obj, self._key)
        return self.__setattr__(key, obj)

    def is_empty(self) -> bool:
        """Is this object empty?."""
        return len(self.as_list()) == 0

    def __iter__(self):
        """Iterator."""
        return iter(self._args)
